use std::{
    collections::VecDeque,
    path::Path,
    sync::{Arc, LazyLock},
    time::Duration,
};

use s2n_quic::{
    Connection, Server,
    stream::{ReceiveStream, SendStream},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{RwLock, broadcast},
    time::sleep,
};

use crate::ServerConfig;
const BUFFER_LENGTH: usize = protocol::BUFFER_LENGTH;

#[derive(Debug)]
struct User {
    sender: broadcast::Sender<f32>,
}
impl User {
    async fn new() -> broadcast::Sender<f32> {
        let (sender, _) = broadcast::channel(BUFFER_LENGTH);
        let new_user = Self {
            sender: sender.clone(),
        };
        ONLINE_USERS.write().await.push(new_user);
        if let Err(err_val) = NEW_USER_NOTIFIER.send(sender.clone()) {
            eprintln!("Error: Send New User | Local | {}", err_val);
        }
        sender.clone()
    }
}
static ONLINE_USERS: LazyLock<RwLock<Vec<User>>> = LazyLock::new(|| vec![].into());
static NEW_USER_NOTIFIER: LazyLock<broadcast::Sender<broadcast::Sender<f32>>> =
    LazyLock::new(|| broadcast::channel(BUFFER_LENGTH).0);

pub async fn start(server_config: ServerConfig) {
    let mut server = Server::builder()
        .with_io(server_config.server_address.as_str())
        .unwrap()
        .with_tls((
            Path::new(&server_config.certificate_path),
            Path::new(&server_config.certificate_key_path),
        ))
        .unwrap()
        .start()
        .unwrap();

    while let Some(connection) = server.accept().await {
        tokio::spawn(handle_client(connection));
    }
}

async fn handle_client(mut connection: Connection) {
    println!(
        "Server Name = {}",
        connection.server_name().unwrap().unwrap().to_string()
    );

    println!(
        "Client Address = {}",
        connection.remote_addr().unwrap().to_string()
    );

    let stream = connection
        .accept_bidirectional_stream()
        .await
        .unwrap()
        .unwrap();
    let (receive_stream, send_stream) = stream.split();

    let new_user_sender = User::new().await;

    tokio::spawn(receiver_audio_data(receive_stream, new_user_sender));
    tokio::spawn(send_audio_data(send_stream));
}

async fn mixer_with_buffer(
    unmixed_audio_data: Arc<RwLock<VecDeque<Arc<RwLock<VecDeque<f32>>>>>>,
) -> Vec<f32> {
    let mut mixed_audio_buffer = vec![];
    for audio_buffer in unmixed_audio_data.read().await.iter() {
        let mut inner_audio_buffer = vec![];

        while let Some(audio_data) = audio_buffer.write().await.pop_front() {
            inner_audio_buffer.push(audio_data);
        }

        for (i, audio_data) in inner_audio_buffer.iter().enumerate() {
            match mixed_audio_buffer.get(i) {
                Some(current_audio_data) => mixed_audio_buffer[i] = current_audio_data + audio_data,
                None => mixed_audio_buffer.push(*audio_data),
            }
        }
    }
    mixed_audio_buffer
}

// async fn mixer_with_channel(
//     unmixed_audio_data: Arc<RwLock<VecDeque<Arc<RwLock<VecDeque<f32>>>>>>,
//     sender: broadcast::Sender<f32>,
// ) {
//     loop {
//         sleep(Duration::from_nanos(10)).await;
//         let mut mixed_audio_buffer = vec![];
//         for audio_buffer in unmixed_audio_data.read().await.iter() {
//             let mut inner_audio_buffer = vec![];
//
//             while let Some(audio_data) = audio_buffer.write().await.pop_front() {
//                 inner_audio_buffer.push(audio_data);
//             }
//
//             for (i, audio_data) in inner_audio_buffer.iter().enumerate() {
//                 match mixed_audio_buffer.get(i) {
//                     Some(current_audio_data) => {
//                         mixed_audio_buffer[i] = current_audio_data + audio_data
//                     }
//                     None => mixed_audio_buffer.push(*audio_data),
//                 }
//             }
//         }
//         for audio_data in mixed_audio_buffer {
//             if let Err(err_val) = sender.send(audio_data) {
//                 eprintln!("Error: Send Mixed Audio | Local | {}", err_val);
//                 return;
//             }
//         }
//     }
// }

async fn new_user_handler(
    mut receiver: broadcast::Receiver<f32>,
    unmixed_audio_data: Arc<RwLock<VecDeque<Arc<RwLock<VecDeque<f32>>>>>>,
) {
    let local_audio_buffer = Arc::new(RwLock::new(VecDeque::new()));
    unmixed_audio_data
        .write()
        .await
        .push_back(local_audio_buffer.clone());

    loop {
        match receiver.recv().await {
            Ok(received_audio_data) => {
                local_audio_buffer
                    .write()
                    .await
                    .push_back(received_audio_data);
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Local | {}", err_val);
                return;
            }
        }
    }
}

async fn send_audio_data(mut send_stream: SendStream) {
    let unmixed_audio_data = Arc::new(RwLock::new(VecDeque::new()));
    for online_user in ONLINE_USERS.read().await.iter() {
        let receiver = online_user.sender.subscribe();
        let unmixed_audio_data = unmixed_audio_data.clone();
        tokio::spawn(new_user_handler(receiver, unmixed_audio_data));
    }
    let unmixed_audio_data_for_handling_new_user = unmixed_audio_data.clone();
    tokio::spawn(async move {
        loop {
            match NEW_USER_NOTIFIER.subscribe().recv().await {
                Ok(new_user_sender) => {
                    let receiver = new_user_sender.subscribe();
                    let unmixed_audio_data = unmixed_audio_data_for_handling_new_user.clone();
                    tokio::spawn(new_user_handler(receiver, unmixed_audio_data));
                }
                Err(err_val) => {
                    eprintln!("Error: Receive New User | Local | {}", err_val);
                    return;
                }
            }
        }
    });

    // let (mixer_sender, mut mixer_receiver) = broadcast::channel(BUFFER_LENGTH);
    // tokio::spawn(mixer_with_channel(unmixed_audio_data, mixer_sender));
    // loop {
    //     match mixer_receiver.recv().await {
    //         Ok(mixed_audio_data) => {
    //             if let Err(err_val) = send_stream.write_f32(mixed_audio_data).await {
    //                 eprintln!("Error: Send Audio Data | Remote | {}", err_val);
    //                 return;
    //             }
    //         }
    //         Err(err_val) => {
    //             eprintln!("Error: Receive Mixed Audio Data | Local | {}", err_val);
    //             return;
    //         }
    //     }
    // }

    loop {
        sleep(Duration::from_millis(50)).await;
        let mixed_audio_buffer = mixer_with_buffer(unmixed_audio_data.clone()).await;
        for mixed_audio_data in mixed_audio_buffer {
            if let Err(err_val) = send_stream.write_f32(mixed_audio_data).await {
                eprintln!("Error: Send Audio Data | Remote | {}", err_val);
                return;
            }
        }
    }
}

async fn receiver_audio_data(mut receive_stream: ReceiveStream, sender: broadcast::Sender<f32>) {
    loop {
        match receive_stream.read_f32().await {
            Ok(received_audio_data) => {
                if let Err(err_val) = sender.send(received_audio_data) {
                    eprintln!("Error: Send Audio Data | Local | {}", err_val);
                    return;
                }
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Remote | {}", err_val);
                return;
            }
        }
    }
}
